/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MOTIFSEARCH_H_
#define MOTIFSEARCH_H_

#include <BALL/FORMAT/PDBFile.h>
#include <BALL/MATHS/analyticalGeometry.h>
#include <BALL/DATATYPE/hashGrid.h>
#include <BALL/COMMON/limits.h>
#include <BALL/SYSTEM/sysinfo.h>
#include <BALL/SYSTEM/timer.h>
#include <BALL/STRUCTURE/numericalSAS.h>
#include <BALL/STRUCTURE/defaultProcessors.h>
#include <BALL/STRUCTURE/secondaryStructureProcessor.h>
#include <BALL/MATHS/matrix44.h>

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>

#include <math.h>

#include "pocketFinder.h"
#include "motif.h"
#include "parameters.h"
#include "ranking.h"
#include "motifHit.h"




using namespace std;
using namespace BALL;
using namespace boost;

struct pair_label_t {
  typedef vertex_property_tag kind;
};
typedef property<pair_label_t, unsigned int> PairLabelProperty;
typedef adjacency_list<listS, vecS, undirectedS, PairLabelProperty> Graph;
typedef Graph::vertex_iterator VertexIt;
typedef property_map<Graph, pair_label_t>::type LabelMap;
typedef Graph::out_edge_iterator EdgeIterator;

/**
 * Class that contains all necessary variables and functions to search one protein structure
 * for suitable attachment site of the enzymatic motif.
 */
class MotifSearch
{

private:


		// private methods

	/**
	 * compute the probabilities for each geometric value in vector geom_values
	 * (i.e. Ca- distance and angles phi, psi, tau).
	 * probability here means similarity of the residue pair that led to geom_values
	 * to a residue pair in the motif (using the average values
	 * and the standard deviations of the geometric values in the motif)
	 */
	vector<double> get_probabilities (vector<double> const &geom_values);

	/**
	 * determines which residues are integral in forming secondary structures (i.e. are in the central part
	 * of a secondary structure as predicted by DSSP)
	 * these residues are not eligible as attachment sites for motif residues, as a change
	 * of the amino acid type could destroy the secondary structure
	 */
	HashSet<Residue*> get_secondary_structure_residues();

	/**
	 * Calculate a vector of residue pairs whose Ca distance is
	 * smaller than the value of the max_distance variable
	 * and larger than the value of the min_distance variable
	 */
	void get_residue_pairs (HashSet<Residue*> const &pocket_residues, HashSet<Residue*> const &secondary_structure_residues);

	/**
	 * compute all residue pairs in the structure whose CA/CB geometries are geometrically similar
	 * (with a probability higher than a threshold) of the residue pairs in the motif
	 */
	void compute_pair_hits ();

	bool determine_pair_hit_rotamers(unsigned int i, unsigned int j, bool swap = false);


	/**
	 * recursion, looks for cliques in the compatibility graph
	 * whose sizes equal the number of residue pairs in the motif.
	 * these cliques represent a complete motif hit (i.e., a residue in the structure is found
	 * for each residue in the motif, and these residues have a CA/CB geometry similar to the ones in the motif)
	 * it tests if the node v can be added to the clique and does that if so
	 */
	void clique_search (const Graph& graph,
	                    const LabelMap& labelMap,
					    vector<unsigned int>& clique,
					    unsigned int v,
					    vector<vector<unsigned int> >& complete_cliques);


	/**
	 * searches for complete motif hits (a set of residues in the structure with one residue for each
	 * resiude of the motif, that have similar CA/CB geometries as the motif).
	 * This is done by creating a compatibility graph for the pair hits and
	 * looking for cliques in it, the cliques must contain as many nodes as there are residue
	 * pairs in the motif
	 */
	void search_complete_motifs (vector<unsigned int> const &sorted_motif_indices);


	/**
	 * checks if there is an atom that fits the BALL expression of an interaction center
	 * closer than the cutoff radius around the reconstructed center position. the center position
	 * is recontructed using the reconstruction matrix
	 */
	void check_interaction_centers ();


	/**
	 * checks complete motif hits if the relative SAS areas of the anchor residues
	 * are similar to the SAS wanted for the motif residues,
	 * and compute a penalty according to the discrepancy
	 */
	void compute_relative_SAS_score ();

	/**
	 * Checks if the sidechain atoms of the constructed motif residues
	 * clash with any backbone atoms of the protein.
	 * clashes are penalized according to the severity of the clash
	 * ( two atoms clash if their distance is smaller than their vdW-radii. the penalty
	 * depends on how much smaller the distance is
	 */
	void check_backbone_clashes ();

	/**
	 * recontructs the positions of all spheres for each motif hit,
	 * and checks of there are any backbone atoms inside the spheres.
	 * if so, computes a penalty.
	 */
	void check_ligand_clashes ();


	/***************************************************************************************/
		//member objects & variables

	/**
	 * motif object defining the geometry, rel. SAS areas, backbone-atome free spheres and
	 * interacting atoms to look for
	 */
	Motif& motif;

	/**
	 * contains pointers to all residues in the current protein that are amino acids
	 */
	vector<Residue*> all_residues;

	/**
	 *  contains the indices of residue pairs in all_residues, the residue pairs
	 *  CA-distances are inside the allowed interval defined by the motif
	 */
	vector< pair <unsigned int, unsigned int> > pair_vector;



	/**
	 * the current PDB structure
	 */
	System& system;

	/**
	 * hits for residue pairs that have a probability score higher than the threshold
	 * pair_hits[i][j] holds the j'th pair hit for motif residue pair i
	 */
	vector<vector<pair<unsigned int, unsigned int> > > pair_hits;

	/**
	 * indices of rotamers that reproduce the pairwise geometry of the catalytic atoms
	 * of the motif residue pair sufficiently well on attachment sites indicated by the pair hit
	 * pair_hit_rotamers[i][j][k] holds the k'th rotamer pair for the j'th pair hit for motif residue pair i
	 */
	vector<vector<vector<pair<unsigned int,unsigned int> > > > pair_hit_rotamers;

	/**
	 * pair_hit_rotamer_scores[i][j][k] holds the partial catalytic atoms geometry penalty score for the
	 * pair hit rotamer pair indicated by pair_hit_rotamers[i][j][k]
	 */
	vector<vector<vector<double> > > pair_hit_rotamer_scores;


	/**
	 *  the probability scores corresponding to the pair_hits entries
	 */
    vector<vector<double> > pair_hit_probabilities;

	/**
	 * contains a MotifHit object for each motif hit
	 */
	list<MotifHit> motif_hits;


	/**
	 * Vectors that hold the lower and upper corner for all the hash grids
	 * so we can be shure that all atoms iof the protein structure are contained within the
	 * grid borders
	 */
	Vector3 lower, upper;

	/**
	 * HashGrid with pointers to all backbone atoms (including Cbeta-Atoms) for the backbone & ligand
	 * clash penalties computation.
	 */
	HashGrid3<Atom*> backbone_atoms_grid;

	/**
	 * provides all global variables for the scaffold selection search
	 */
	Parameters parameters;


	/////////////////////////////////////////////////////////////////////////////////////

public:

	// constructors

	/**
	 * this contructor assigns its arguments to the respective
  	 * member variables of the Motifsearch
  	 */
	MotifSearch (System& pdb_system,
				 Motif& motif_ref,
				 Parameters const &parameters_ref);

	/**
	 * destructor
	 */
	virtual ~MotifSearch ();

		// public methods


	/**
	 * searches for possible motif geometries the PDB System
	 * computes the scores for these hits and adds them to the ranking datastructure
	 * returns the number of found motif hits
	 */
	unsigned int search_for_motifs (Ranking& ranking);


};

#endif /*MOTIFSEARCH_H_*/
