/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * uses code from LIGSITECS that was modified & adapted for ScaffolSelection
 * by Christoph Malisi under the GPL Gnu Public License
 *
 * The LIGSITECS Copyright notice:
 *      LIGSITECS, a pocket identification method using Connolly surface
 *      http://scoppi.biotec.tu-dresden.de
 *
 *      written by
 *      Bingding Huang
 *      Bioinformatics group
 *      Biotec & Department of Computing
 *      Tazberg 47, 01307
 *      TU Dresden, Germany
 *      Email bhuang@biotec.tu-dresden.de
 *
 *      Copyright (C) LIGSITECS  Bingding Huang
 */

#ifndef POCKETFINDER_H_
#define POCKETFINDER_H_


#include <BALL/KERNEL/system.h>
#include <BALL/FORMAT/PDBFile.h>
#include <BALL/DATATYPE/string.h>
#include <BALL/DATATYPE/hashMap.h>
#include <BALL/STRUCTURE/defaultProcessors.h>
#include <BALL/STRUCTURE/analyticalSES.h>
#include <BALL/STRUCTURE/geometricProperties.h>
#include <BALL/STRUCTURE/geometricTransformations.h>
#include <BALL/STRUCTURE/surfaceProcessor.h>
#include <BALL/STRUCTURE/numericalSAS.h>
#include <BALL/MATHS/surface.h>
#include <BALL/MATHS/vector4.h>
#include <BALL/SYSTEM/timer.h>
#include <BALL/COMMON/exception.h>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/connected_components.hpp>
#include <boost/config.hpp>

#include <iostream>
#include <set>
#include <stdlib.h>
#include "parameters.h"


using namespace BALL;
using namespace std;

/**
 * Contains the functionalities to search for ligand binding pockets in a PDB structure
 * using the LIGSITEcs algorithm by Huang 2006
 */
class PocketFinder {
public:

	/**
	 * Constructor, sys is the PDB system in which to look for pockets
	 */
	PocketFinder(System & sys, Parameters const &params);

	/**
	 * Destructor
	 */
	virtual ~PocketFinder();

	/**
	 * determines the residues in the PDB structures that form pockets
	 */
	HashSet<Residue*> get_pocket_residues();

private:

	/**
	 * contains the global ScaffoldSelection user-defined parameters,
	 * including the LIGSITEcs parameters
	 */
	Parameters parameters;

	/**
	 * The PDB systemin which to look for pockets
	 */
	System& system;

	/**
	 * providing all necessary information for a grid point of the LIGSITEcs pocket search grid
	 */
	struct GridPoint
	{
		/**
		 * the index of the grid point, i.e. coordinates of unsigned int values
		 */
		Vector3 index;

		/**
		 * the position of the grid point in the PDB system
		 */
		Vector3 position;

		/**
		 * grid index ???
		 */
		int grid_index;

		/**
		 * ??
		 */
		bool inside;

		/**
		 * true if the grid point lays inside the protein boundaries
		 */
		bool interactWithProtein;

		/**
		 * true if the grid point is on or very near to the protein surface
		 */
		bool onSurface;

		/**
		 * ??
		 */
		int num_vertice;


	};

	/**
	 * Represents a pocket on the protein surface
	 */
	class Pocket
	{

	public:

		/**
		 * default constructor
		 */
		Pocket();

		/**
		 * Constructor
		 */
		Pocket(int number, Vector3 pos, int ind);

		/**
		 * Destructor
		 */
		virtual ~Pocket();

		/**
		 * compares two pockets by comparing the number of the grid points contributing to the pockets
		 */
		bool operator < (const Pocket& p) const;

		/**
		 * returns the position of the pocket (i.e. the position of one of it's grid points)
		 */
		Vector3 getPosition();

		/**
		 * set the pocket position
		 */
		void setPosition(Vector3 pos);

		/**
		 * set the pocket index
		 */
		void setIndex(int n);

		/**
		 * get the pocket index
		 */
		const int getIndex() const;

		/**
		 * add a neighboring pocket
		 */
		void addNeighbour(int n);

		/**
		 * get the set of neighboring pockets (i.e. their indices in a vector of pockets)
		 */
		HashSet<int> getNeighbour() const;

		/**
		 * return the pocket size (i.e. the number of contributing grid points)
		 */
		const int getSize() const;

		/**
		 * remembers that the pocket is added to a cluster of pockets, i.e. a neighborhood of another pocket
		 */
		void setAdded();

		/**
		 * checks if the pocket is added to a cluster
		 */
		bool isAdded();

	private:
		/**
		 * number of SSS events of the pocket's "root" grid point??
		 */
		int n;

		/**
		 * position of the pocket's "root" grid point
		 */
		Vector3 position;

		/**
		 * index of the pocket
		 */
		int index;

		/**
		 * the indices of the neighboring pockets, i.e. the pockets in the same cluster
		 */
		HashSet<int> neighbour;

		/**
		 * true if the pocket is added to a cluster
		 */
		bool added;

	};

	/**
	 * computes the center of mass of the protein
	 */
	Vector3 getMassCenter();

	/**
	 * return the extent of the protein (i.e. the largest absolute values of all three coordinates)
	 */
	Vector3 getExtent();

	/**
	 * returns the origin of the pocket grid
	 */
	Vector3 getGrid_origin(Vector3 GRID, float GRID_SPACING);

};


#endif /*POCKETFINDER_H_*/
