/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "motifHit.h"
using namespace BALL;
using namespace std;


MotifHit::MotifHit(String const &system_name,
                   vector<Residue*> const &res_vec,
                   Motif& mot) :
	is_wt_hit(),
	pair_rotamer_scores(),
	allowed_rotamer_sets(),
	indices(),
	rotamer_indices(),
	rotamer_probability(),
	catalytic_atoms_geometry_penalty(),
	relative_SAS_penalty(),
	backbone_clash_penalty(),
	ligand_clash_penalty(),
	interaction_centers_penalty(),
	mapping_matrix(),
	pdb_name( system_name),
	all_residues( res_vec),
	interaction_atoms(),
	motif( mot)
{
	unsigned int i;
	for (i = 0; i < motif.get_residue_count(); i++)
	{
		indices.push_back(i);
		set<unsigned int> tmp;
		allowed_rotamer_sets.push_back(tmp);
	}
	for (i = 0; i < motif.get_residue_pair_count(); ++i)
	{
		HashMap<unsigned int, double> tmp;
		pair_rotamer_scores.push_back(tmp);
	}
}

MotifHit::MotifHit() :
	is_wt_hit(),
	pair_rotamer_scores(),
	allowed_rotamer_sets(),
	indices(),
	rotamer_indices(),
	rotamer_probability(),
	catalytic_atoms_geometry_penalty(),
	relative_SAS_penalty(),
	backbone_clash_penalty(),
	ligand_clash_penalty(),
	interaction_centers_penalty(),
	mapping_matrix(),
	pdb_name(),
	all_residues(),
	interaction_atoms(),
	motif()
{
}

MotifHit::~MotifHit()
{
}


/*
 * checks every complete motif hit (that was computed using the distribution of geometric values):
 * checks each combination of rotamers of the motif residues constructed on the backbone of the
 * target protein at the positions indicated by the motif hit. the combination that minimizes
 * the sum of the rmsd of the differences of distances between the catalytic atoms and the rmsd
 * of differences between torsion angles of catalytic atoms will be used as the reconstructed motif hit
 */
bool MotifHit::construct_motif_rotamers (HashGrid3<Atom*> const &backbone_atoms_grid, double backbone_clash_threshold)
{
	// get the positions of 2 backbone atoms and CB for each motif hit residue, so we can
	// calculate a transformation matrix that maps the motif rotamers onto the backbone
	// at these positions
	Vector3 CA, CB, C, CA_rot, CB_rot, C_rot;

	for (unsigned int i = 0; i < motif.get_residue_count(); ++i)
	{
		CA = all_residues[indices[i]]->getAtom("CA")->getPosition();
		C = all_residues[indices[i]]->getAtom("C")->getPosition();
		if (all_residues[indices[i]]->getName() != "GLY")
		{
			CB = all_residues[indices[i]]->getAtom("CB")->getPosition();
		}
		else
		{
			CB = scsel_util::get_ideal_cb_position_for_gly(all_residues[indices[i]]->getAtom("N")->getPosition(), C, CA);
		}

		// get positions of the backbone atoms of the current rotamers
		CA_rot = motif.get_rotamer(i, 0)->getAtom("CA")->getPosition();
		CB_rot = motif.get_rotamer(i, 0)->getAtom("CB")->getPosition();
		C_rot = motif.get_rotamer(i, 0)->getAtom("C")->getPosition();
		Matrix4x4 T;
		StructureMapper	mapper;

		// compute a transformation that maps the current rotamer set onto the backbone at the
		// position specified by the motif hit
		T = mapper.matchPoints (CA_rot, CB_rot, C_rot, CA, CB, C);
		mapper.setTransformation(T);

		// loop over all possible rotamers for the current motif residue
		for (unsigned int j = 0; j < motif.get_number_of_rotamers_for_residue(i); ++j)
		{
			motif.get_rotamer (i, j)->apply(mapper);
		}
	}

	unsigned int clash_rotamers = 0;
	double violation = 0.0;
	Expression sidechain_expression("!backbone() AND !element(H)");
	// sort out rotamers that clash with the backbone
	for (unsigned int i = 0; i < motif.get_residue_count(); ++i)
	{
		set<unsigned int>::iterator rit = allowed_rotamer_sets[i].begin();
		while (rit != allowed_rotamer_sets[i].end())
		{
			Residue* current_rotamer = motif.get_rotamer(i, *rit);

			double clash_penalty = 0.0;
			// loop over all atoms in the current constructed motif residue
			AtomIterator atom_it = current_rotamer->beginAtom();
			for (; atom_it != current_rotamer->endAtom(); ++ atom_it)
			{
				// only look at non-backbone atoms and non-hydrogens
				if (sidechain_expression(*atom_it))
		    	{
		    		// get the hashgrid box that contains the position of the current atom
		    		const HashGridBox3<Atom*>* box = backbone_atoms_grid.getBox(atom_it->getPosition());
					Position x, y, z;
					backbone_atoms_grid.getIndices(*box, x, y, z);

		    		// loop over the neighbour boxes
		    		for (Index xi = -1; xi <= 1; xi++)
					{
						const Position nx = x + xi;
						for (Index yi = -1; yi <= 1; yi++)
						{
							const Position ny = y + yi;

							for (Index zi = -1; zi <= 1; zi++)
							{
								box = backbone_atoms_grid.getBox(nx, ny, z+zi);

								// if there are no atoms in the box, skip it
								if (box->isEmpty())
								{
									continue;
								}

								// loop over all atoms in the current box
								HashGridBox3<Atom*>::ConstDataIterator bb_it = box->beginData();
								for (; bb_it != box->endData() ; ++bb_it)
								{
									// get the parent residue of the current backbone atom
									Residue* res = (*bb_it)->getResidue();

									// if the current atom is from the current motif residue position in the hit
									// or its neighbour, ignore it
									unsigned int index = indices[i];
									if ( res != all_residues[index] &&
										 (index > 0 && res !=  all_residues[index - 1]) &&
										 (index < all_residues.size() - 1 && res !=  all_residues[index + 1]) )
									{
										// if the distance between the two atoms is too small,
										// register a clash

										violation =   ((*bb_it)->getPosition() - atom_it->getPosition()).getLength()
											 - ((*bb_it)->getElement().getVanDerWaalsRadius()
											                   + atom_it->getElement().getVanDerWaalsRadius()) * 0.9 ;
										if (violation < 0.0)
										{
											clash_penalty += pow(violation, 2);
										}
									}
								}
							}
						}
					}
				}
			}
			if (clash_penalty > backbone_clash_threshold)
			{
				clash_rotamers++;
				allowed_rotamer_sets[i].erase(rit++);
			}
			else
			{
				rit++;
			}
		}
		if (allowed_rotamer_sets[i].size() == 0)
		{
			return false;
		}
	}

	// precompute the partial geometric scores for residue pairs,
	// so we do not have to compute the values every time in the recursion
	for (unsigned int i = 0; i < motif.get_residue_pair_count(); ++i)
	{
		pair<unsigned int, unsigned int> motpair = motif.get_residue_pair(i);
		Vector3 catatom11, catatom12, catatom13, catatom21, catatom22, catatom23;
		vector<vector<double> > vec1;

		// loop over all allowed rotamers for the current motif residue pair
		set<unsigned int>::iterator rit1 = allowed_rotamer_sets[motpair.first].begin();
		for (; rit1 != allowed_rotamer_sets[motpair.first].end(); ++rit1)
		{
			// get the positions of the catalytic atoms of the current rotamer for the first residue
			catatom11 = motif.get_rotamer(motpair.first, *rit1)->getAtom(motif.get_catalytic_atom_name(motpair.first, 0))->getPosition();
			catatom12 = motif.get_rotamer(motpair.first, *rit1)->getAtom(motif.get_catalytic_atom_name(motpair.first, 1))->getPosition();
			catatom13 = motif.get_rotamer(motpair.first, *rit1)->getAtom(motif.get_catalytic_atom_name(motpair.first, 2))->getPosition();

			set<unsigned int>::iterator rit2 = allowed_rotamer_sets[motpair.second].begin();
			for (; rit2 != allowed_rotamer_sets[motpair.second].end(); ++rit2)
			{
				if (!pair_rotamer_scores[i].has((*rit1) * 10000 + (*rit2)) )
				{
					// get the positions of the catalytic atoms of the current rotamer for the first residue
					catatom21 = motif.get_rotamer(motpair.second, *rit2)->getAtom(motif.get_catalytic_atom_name(motpair.second, 0))->getPosition();
					catatom22 = motif.get_rotamer(motpair.second, *rit2)->getAtom(motif.get_catalytic_atom_name(motpair.second, 1))->getPosition();
					catatom23 = motif.get_rotamer(motpair.second, *rit2)->getAtom(motif.get_catalytic_atom_name(motpair.second, 2))->getPosition();

					double distance_score = 0.0;
					double torsion_score = 0.0;

					double rot_length = (catatom11 - catatom21).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 0, 0)), 2);
					rot_length = (catatom11 - catatom22).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 0, 1)), 2);
					rot_length = (catatom11 - catatom23).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 0, 2)), 2);
					rot_length = (catatom12 - catatom21).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 1, 0)), 2);
					rot_length = (catatom12 - catatom22).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 1, 1)), 2);
					rot_length = (catatom12 - catatom23).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 1, 2)), 2);
					rot_length = (catatom13 - catatom21).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 2, 0)), 2);
					rot_length = (catatom13 - catatom22).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 2, 1)), 2);
					rot_length = (catatom13 - catatom23).getLength();
					distance_score += pow((rot_length - motif.get_catalytic_atom_distance(motpair.first, motpair.second, 2, 2)), 2);

					double torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom12.x, catatom12.y, catatom12.z, catatom21.x, catatom21.y, catatom21.z, catatom22.x, catatom22.y, catatom22.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 1, 0, 1)), 2);
					torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom12.x, catatom12.y, catatom12.z, catatom21.x, catatom21.y, catatom21.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 1, 0, 2)), 2);
					torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom13.x, catatom13.y, catatom13.z, catatom21.x, catatom21.y, catatom21.z, catatom22.x, catatom22.y, catatom22.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 2, 0, 1)), 2);
					torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom13.x, catatom13.y, catatom13.z, catatom21.x, catatom21.y, catatom21.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 2, 0, 2)), 2);
					torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom12.x, catatom12.y, catatom12.z, catatom22.x, catatom22.y, catatom22.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 1, 1, 2)), 2);
					torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z, catatom13.x, catatom13.y, catatom13.z, catatom22.x, catatom22.y, catatom22.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 0, 2, 1, 2)), 2);
					torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z, catatom13.x, catatom13.y, catatom13.z, catatom21.x, catatom21.y, catatom21.z, catatom22.x, catatom22.y, catatom22.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 1, 2, 0, 1)), 2);
					torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z, catatom13.x, catatom13.y, catatom13.z, catatom21.x, catatom21.y, catatom21.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 1, 2, 0, 2)), 2);
					torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z, catatom13.x, catatom13.y, catatom13.z, catatom22.x, catatom22.y, catatom22.z, catatom23.x, catatom23.y, catatom23.z);
					torsion_score += pow ((torsion - motif.get_catalytic_atom_torsion (motpair.first, motpair.second, 1, 2, 1, 2)), 2);

					pair_rotamer_scores[i][(*rit1) * 10000 + (*rit2)] = torsion_score + distance_score;
				}
			}
		}
	}

	// in this variable, the smallest deviation of catalytic atom distance differences
	// will be stored
	double rmsd_min = Limits<double>::max();

	// check every possible combination of rotamers for the motif residues at the
	// backbone positions of the motif hit to find the combination with the
	// smallest rmsd
	vector<unsigned int> motif_rotamer_set, minimal_rotamer_set;
	rotamer_construction_recursion( motif_rotamer_set, minimal_rotamer_set, rmsd_min, 0);

	double rotamer_set_probability = 1.0;
	for (unsigned int i=0; i < motif.get_residue_count(); ++i)
	{
		rotamer_set_probability *= motif.get_rotamer_frequency(i, minimal_rotamer_set[i]);
	}

	// remember the best rotamer set for this motif hit and its rmsd
	rotamer_probability = rotamer_set_probability;
	rotamer_indices = minimal_rotamer_set;

	// 18 values for each residue pair were computed and summed (9 catalytic atoms distances
	// and 9 torsion angles)
	// normalises the penalty score sum and takes the squareroot, accoridng to the
	// rmsd formula
	rmsd_min /= 18 * motif.get_residue_pair_count();
	rmsd_min = sqrt(rmsd_min);
	catalytic_atoms_geometry_penalty = rmsd_min;
	return true;
}


/*
 * recursively constructs every rotamer combination on the positions of a
 * motif hit, and identifies the rotamer set that has the most similar
 * geometry to the original motif.
 * similarity is measured by the difference of the distances between the
 * catalytic atoms of the motif residues/rotamers and the difference of
 * torsion angles of any 2 catayltic atoms of any 2 residues between the motif
 * residues and the built rotamers
 */
void MotifHit::rotamer_construction_recursion (vector<unsigned int>& motif_rotamer_set,
							   				   vector<unsigned int>& minimal_rotamer_set,
							   				   double& minimal_geometry_penalty,
							   				   double partial_score)
{
	// found complete set of rotamers (for every motif residue one)
	if(motif_rotamer_set.size() == motif.get_residue_count())
	{
		// if the current rotamer set leads to a geometry more similar to the original motif (in terms of
		// distances and torsions of catalytic atoms), then remember it as the best rotamer set
		if (partial_score < minimal_geometry_penalty)
		{
			minimal_geometry_penalty = partial_score;
			minimal_rotamer_set.clear();
			minimal_rotamer_set.insert (minimal_rotamer_set.end(), motif_rotamer_set.begin(), motif_rotamer_set.end() );
		}
		return;
	}

	// there are not yet enough rotamers in the set, so loop over the rotamers for the next
	// motif residue and recurse for each of them
	else
	{
		// how "deep" are we in the recursion tree? this is also the index of the motif residue
		// that is examined in this recursion instance
		unsigned int depth = motif_rotamer_set.size(), k = 0;

		map<double, unsigned int> partial_scores;

		// loop over all allowed rotamers for the current motif residue
		set<unsigned int>::iterator rot_it = allowed_rotamer_sets[depth].begin();
		for (; rot_it != allowed_rotamer_sets[depth].end(); ++rot_it)
		{


			// if the recursion is 0, there are no rotamers yet in the set, and no partial penalty scores
			// need to be computed as they are all defined for residue pairs.
			// Else compute the partial penalty scores that each possible rotamer for this depth produces
			// with the rotamers already in the set
			if (depth > 0)
			{
				// the current partial geometry penalty score is the partial penalty score of the rotamers
				// already in the current set, we add to it the penalty scores of the current rotamer for
				// the residue with index depth
				double current_partial_score = 0;

				//loop over all rotamers already in the motif rotamer set
				for (unsigned int j = 0; j < depth; ++j)
				{
					current_partial_score += pair_rotamer_scores[motif.get_residue_pair_index(j, depth)][motif_rotamer_set[j]*10000 + (*rot_it)];
    			}
				double partial_iter_score = current_partial_score + partial_score;

				partial_scores[partial_iter_score] = *rot_it;
			}
			++k;
		}

		// if the depth is > 0, we have computed partial penalty scores for each
		// rotamer for the residue at this depth
		if ( depth > 0)
		{
			// loop over the rotamers in the order of the partial penalty scores,
			// small penalty scores first
			map<double, unsigned int>::iterator it = partial_scores.begin();
			for (; it != partial_scores.end(); ++it)
			{
				// if the partial penalty score is smaller than the current minimal complete penalty score, the
				// corresponding rotamer could be part of the global minimal penalty geometry
				if (it->first < minimal_geometry_penalty)
				{
					motif_rotamer_set.push_back(it->second);
					rotamer_construction_recursion (motif_rotamer_set, minimal_rotamer_set,
													minimal_geometry_penalty, it->first);
					motif_rotamer_set.pop_back();
				}
				// if the partial penalty score is larger than the current minimal penalty score, we
				// do not need to recurse further, because it cannot be part of the minimal penalty geometry
				// because the rotamers are sorted by their partial scores, we do not need to recurse for
				// the remaining rotamers either
				// in effect, we cut away a part of the rotamer search tree that we can be sure of doesn't
				// contain the best geometry
				else break;
			}
		}
		else
		{
			set<unsigned int>::iterator rot_it = allowed_rotamer_sets[depth].begin();
			for (; rot_it != allowed_rotamer_sets[depth].end(); ++rot_it)
			{
				motif_rotamer_set.push_back(*rot_it);
				rotamer_construction_recursion (motif_rotamer_set, minimal_rotamer_set,
												minimal_geometry_penalty, 0);
				motif_rotamer_set.pop_back();
			}
		}
	}
}


/*
 * Computes for each motif hit a transformation matrix that maps the coordinates of the motif catalytic atoms
 * onto the coordinates of these atoms in the reconstructed motif (according to the hit)
 */
void MotifHit::compute_mapping_matrix ()
{

	Matrix4x4 T;

	StructureMapper	mapper;

	vector<Vector3> hit_atom_positions;

	// loops over the rotamers of the minimal rotamer set for each motif residue
	for (unsigned int j = 0; j < motif.get_residue_count(); ++j)
	{
		// get the built rotamer for the current motif residue
		Residue* rotamer = motif.get_rotamer(j, rotamer_indices[j]);


		// get the residue in the current protein whose sidechain will be replaced
		// by the motif residue
		Residue* anchor = all_residues[indices[j]];

		// compute a transformation from the backbone positions of the motif rotamers to the
		// backbone positions of the motif hit
		T = StructureMapper::matchBackboneAtoms(*rotamer, *anchor);
/*		T = matchPoints (rotamer->getAtom("CA")->getPosition(),
								rotamer->getAtom("CB")->getPosition(),
								rotamer->getAtom("C")->getPosition(),
								anchor->getAtom("CA")->getPosition(),
								CB_anchor,
								anchor->getAtom("C")->getPosition());
*/


		// map the motif residue onto its position in the protein
		// according to the motif hit
		mapper.setTransformation(T);
		rotamer->apply(mapper);

		// remember the positions of the catalytic atoms for the RMSD minimization
		hit_atom_positions.push_back(rotamer->getAtom(motif.get_catalytic_atom_name(j, 0))->getPosition());
		hit_atom_positions.push_back(rotamer->getAtom(motif.get_catalytic_atom_name(j, 1))->getPosition());
		hit_atom_positions.push_back(rotamer->getAtom(motif.get_catalytic_atom_name(j, 2))->getPosition());
	}

	// compute the reconstruction matrix using BALL's RMSDMinimizer
	RMSDMinimizer::Result r = RMSDMinimizer::computeTransformation(motif.get_catalytic_atom_positions_vector(),
																   hit_atom_positions);
	mapping_matrix = r.first;
}


/*
 * checks if there is an atom that fits the BALL expression of an interaction center
 * closer than the cutoff radius around the reconstructed center position. the center position
 * is recontructed using the reconstruction matrix
 */
bool MotifHit::check_interaction_centers (HashGrid3<Atom*> const &interaction_atoms_grid, bool force_search_all)
{
	Expression backbone_expression("backbone()");

	double penalty = 0.0;
	bool center_exists = false;

	interaction_atoms.clear();

	// loop over all interaction centers
	for (unsigned int i = 0; i < motif.get_interaction_center_count(); ++i)
	{
		// compute the position of the interaction center relative to the current motif hit by
		// multiplicating the original position with the reconstruction matrix of the hit
		Vector3 reconstructed_center = mapping_matrix * motif.get_interaction_center_position(i);
		center_exists = false;
		Atom* interaction_atom = NULL;

		// this is used to store the minial distance of the reconstructed center position to an atom that fits
		// the interaction center BALL expression
		double min_distance = Limits<float>::max();

		// get the hashgrid box that contains the position of the current interaction center
		const HashGridBox3<Atom*>* box = interaction_atoms_grid.getBox(reconstructed_center);

		Position x, y, z;
		interaction_atoms_grid.getIndices(*box, x, y, z);

		// loop over the neighbour boxes
		for (Index xi = -1; xi <= 1; xi++)
		{
			const Position nx = x + xi;
			for (Index yi = -1; yi <= 1; yi++)
			{
				const Position ny = y + yi;
				for (Index zi = -1; zi <= 1; zi++)
				{
					box = interaction_atoms_grid.getBox(nx, ny, z+zi);
					if (!box)
					{
						continue;
					}
					// if there are no atoms in the box, skip it
					if (box->isEmpty())
					{
						continue;
					}
					// loop over all atoms in the current box
					HashGridBox3<Atom*>::ConstDataIterator bb_it = box->beginData();
					for (; bb_it != box->endData() ; ++bb_it)
					{
						// if the current atom fits the current interaction center BALL expression
						if (motif.get_interaction_center_expression(i)(**bb_it))
						{
							// compute the distance between the potential interaction atom and the reconstructed ideal center position
							double distance = ( (*bb_it)->getPosition() - reconstructed_center).getLength();

							// if the distance is smaller than the cutoff and if this is the first potential interaction atom
							// or th distance is the smallest until now, go on
							if (distance < motif.get_interaction_center_cutoff_radius(i)
								&& (!center_exists || distance < min_distance) )
							{
								// check if the potiential interaction atom is a sidechain atom of an anchor residue
								// (i.e. a residue whose sidechain will be removed by a motif residue) if that is so,
								// the pot. atom cannot be used as an interaction center
								// if it is a backbone atom, it will stay even if the sidechain is altered, so it can be used as
								// an interaction center
								bool is_anchor = false;
								if (!backbone_expression(**bb_it) )
								{
									for (unsigned int k = 0; k < motif.get_residue_count(); ++k)
									{
										if ( (*bb_it)->getResidue() == all_residues[indices[k]] )
										{
											is_anchor = true;
											break;
										}
									}
								}
								// the potential interaction atom will be used as a interaction center
								if (!is_anchor)
								{
									center_exists = true;
									min_distance = distance;
									interaction_atom = (*bb_it);
								}
							}
						}
					}
				}
	    	}
		}
		interaction_atoms.push_back(interaction_atom);
		// if for one of the interaction centers we could not find an suitable atom,
		// stop the search for this motif hit
		if (!center_exists)
		{
			if (!force_search_all) break;
		}
		// the interaction center penalty is the squared difference between the reconstructed ideal position
		// and the actual position of the found interaction atom
		else
		{
			penalty += pow(min_distance, 2);
		}
	}
	// if for one interaction center there is no suitable atom in the structure,
	// delete the motif hit
	if (!center_exists)
	{
		return false;
	}
	// if there is a suitable interaction atom for each interaction center, remember the penalty for
	// the position differences
	interaction_centers_penalty = penalty;
	return true;
}


/**
 * Checks if the sidechain atoms of the constructed motif residues
 * clash with any backbone atoms of the protein.
 * clashes are penalized according to the severity of the clash
 * ( two atoms clash if their distance is smaller than their vdW-radii. the penalty
 * depends on how much smaller the distance is
 */
void MotifHit::check_backbone_clashes (HashGrid3<Atom*> const &backbone_atoms_grid)
{
	Expression sidechain_expression("!backbone() AND !element(H)");
	Matrix4x4 T;
	StructureMapper	mapper;

	// helper variable to temporary store the violation of minimal vdW distance
	double violation;

	// stores pointers to the built rotamers of the motif hit that are mapped onto the
	// backbone positions of the hit
	vector<Residue*> rotamer_set;

	double clash_penalty = 0.0;

	// loops over the rotamers of the minimal rotamer set for each motif residue
	for (unsigned int j = 0; j < motif.get_residue_count(); ++j)
	{
		// get the built rotamer for the current motif residue
		Residue* rotamer = motif.get_rotamer(j, rotamer_indices[j]);

		// get the residue in the current protein whose sidechain will be replaced
		// by the motif residue
		Residue* anchor = all_residues[indices[j]];

		Vector3 CB_anchor, CB_rotamer;
		if(anchor->getName() == "GLY")
		{
			CB_anchor = scsel_util::get_ideal_cb_position_for_gly (anchor->getAtom("N")->getPosition(),
													   anchor->getAtom("C")->getPosition(),
													   anchor->getAtom("CA")->getPosition());
		}
		else
		{
			CB_anchor = anchor->getAtom("CB")->getPosition();
		}
		// compute a transformation matrix that maps the built rotamer onto the backbone
		// of the protein according to the motif hit
		T = mapper.matchPoints (rotamer->getAtom("CA")->getPosition(),
								rotamer->getAtom("CB")->getPosition(),
								rotamer->getAtom("C")->getPosition(),
								anchor->getAtom("CA")->getPosition(),
								CB_anchor,
								anchor->getAtom("C")->getPosition());

		// map the motif residue onto its anchor position in the protein
		// according to the motif hit
		mapper.setTransformation(T);
		rotamer->apply(mapper);
		rotamer_set.push_back(rotamer);
	}


	// loop over all motif residues, now in the positions
	// defined by the current motif hit
	for (unsigned int k = 0; k < motif.get_residue_count(); ++k)
	{
		// loop over all atoms in the current constructed motif residue
		AtomIterator atom_it = rotamer_set[k]->beginAtom();
		for (; atom_it != rotamer_set[k]->endAtom(); ++ atom_it)
		{
			// only look at non-backbone atoms and non-hydrogens
			if (sidechain_expression(*atom_it))
	    	{
	    		// get the hashgrid box that contains the position of the current atom
	    		const HashGridBox3<Atom*>* box = backbone_atoms_grid.getBox(atom_it->getPosition());
				Position x, y, z;
				backbone_atoms_grid.getIndices(*box, x, y, z);

	    		// loop over the neighbour boxes
	    		for (Index xi = -1; xi <= 1; xi++)
				{
					const Position nx = x + xi;
					for (Index yi = -1; yi <= 1; yi++)
					{
						const Position ny = y + yi;

						for (Index zi = -1; zi <= 1; zi++)
						{
							box = backbone_atoms_grid.getBox(nx, ny, z+zi);

							// if there are no atoms in the box, skip it
							if (box->isEmpty())
							{
								continue;
							}

							// loop over all atoms in the current box
							HashGridBox3<Atom*>::ConstDataIterator bb_it = box->beginData();
							for (; bb_it != box->endData() ; ++bb_it)
							{
								// get the parent residue of the current backbone atom
								Residue* res = (*bb_it)->getResidue();

								// if the current atom is from the current motif residue position in the hit
								// or its neighbour, ignore it
								unsigned int index = indices[k];
								if ( res != all_residues[index] &&
									 (index > 0 && res !=  all_residues[index - 1]) &&
									 (index < all_residues.size() - 1 && res !=  all_residues[index + 1]) )
								{
									// if the distance between the two atoms is too small,
									// register a clash

									violation =   ((*bb_it)->getPosition() - atom_it->getPosition()).getLength()
										 - ((*bb_it)->getElement().getVanDerWaalsRadius()
										                   + atom_it->getElement().getVanDerWaalsRadius()) * 0.9 ;
									if (violation < 0.0)
									{
										clash_penalty += pow(violation, 2);
									}
								}
							}
						}
					}
				}
	    	}
		}
	}
	// store the total backbone clash penalty of the motif hit
	backbone_clash_penalty = clash_penalty;
}


/*
 * recontructs the positions of all spheres for the motif hit,
 * and checks of there are any backbone atoms inside the spheres.
 * if so, computes a penalty.
 */
void MotifHit::check_ligand_clashes(HashGrid3<Atom*> const &backbone_atoms_grid)
{
	// helper variable to temporary store the violation of minimal vdW distance
	double violation;
	double clash_penalty = 0.0;

	StructureMapper	mapper;
	mapper.setTransformation(mapping_matrix);
	System *ligand_system = motif.get_substrate_system();
	ligand_system->apply(mapper);

	AtomIterator a_it = ligand_system->beginAtom();
	for (; a_it != ligand_system->endAtom(); ++a_it)
	{
		Vector3 atom_position = a_it->getPosition();
//		cout << a_it->getName() << "  " << atom_position << endl;
		// get the hashgrid box that contains the position of the current atom
		const HashGridBox3<Atom*>* box = backbone_atoms_grid.getBox(atom_position);

		Position x, y, z;
		backbone_atoms_grid.getIndices(*box, x, y, z);

		// loop over the neighbour boxes
		for (Index xi = -1; xi <= 1; xi++)
		{
			const Position nx = x + xi;
			for (Index yi = -1; yi <= 1; yi++)
			{
				const Position ny = y + yi;
				for (Index zi = -1; zi <= 1; zi++)
				{
					box = backbone_atoms_grid.getBox(nx, ny, z+zi);
					if (!box)
					{
						continue;
					}
					// if there are no atoms in the box, skip it
					if (box->isEmpty())
					{
						continue;
					}
					// loop over all atoms in the current box
					HashGridBox3<Atom*>::ConstDataIterator bb_it = box->beginData();
					for (; bb_it != box->endData() ; ++bb_it)
					{

						// if the distance between the atom and the center of the sphere is smaller
						// than the radius, add a penalty
						violation = (((*bb_it)->getElement().getVanDerWaalsRadius() * 0.9)
				                   + (a_it->getElement().getVanDerWaalsRadius() * 0.9))
							 - ((*bb_it)->getPosition() - atom_position).getLength();
						if (violation > 0.0)
						{
							clash_penalty += pow(violation, 2);
						}
					}
				}
	    	}
		}
	}
	// store the sphere clash penalty for the current motif hit
	ligand_clash_penalty = clash_penalty;
}


/*
 * checks complete motif hits, that map a residue of the structure to a residue in the motif, if the SAS
 * of the structure residues are similar to the SAS wanted for the motif residues,
 * and add a penalty according to the discrepancy to the probablity score of the motif hit
 */
bool MotifHit::compute_relative_SAS_score (vector<double> const &residue_SAS_areas)
{
	double penalty = 0.0;

	// loop over all residues in the hit
	for (unsigned int j = 0; j < motif.get_residue_count(); ++j)
	{
		// the penalty factor is the root mean square deviation between the relative SAS of the motif residues
		// to the relative SAS of the residues in the complete motif hit
		if (residue_SAS_areas[indices[j]] < 0.0) return false;
		penalty += pow( (motif.get_relative_SAS_for_residue(j) - residue_SAS_areas[indices[j]]), 2);
	}
	penalty /= motif.get_residue_count();
	penalty = sqrt(penalty);
	relative_SAS_penalty = penalty;
	return true;
}

void MotifHit::insert_allowed_rotamers(unsigned int i, const set<unsigned int>& rotamer_set)
{
	allowed_rotamer_sets[i].insert(rotamer_set.begin(), rotamer_set.end());
}

/**
 * set the residue index at position pos to value i
 */
void MotifHit::set_residue_index(unsigned int pos, unsigned int i)
{
	indices[pos] = i;
}

const Atom* MotifHit::get_interaction_atom (unsigned int i) const
{
	return interaction_atoms[i];
}

Matrix4x4 MotifHit::get_mapping_matrix() const
{
	return mapping_matrix;
}

/**
 * returns a pointer to the Rotamer determined for motif residue no. i attachment
 */
Residue* MotifHit::get_rotamer(unsigned int i)
{
	return motif.get_rotamer(i, rotamer_indices[i]);
}


/**
 * returns the rotamer probability score of this motif hit
 */
const double MotifHit::get_rotamer_probability() const
{
	return rotamer_probability;
}

/**
 * returns the catalytic atoms geometry penalty score of this motif hit
 */
const double MotifHit::get_catalytic_atoms_geometry_penalty() const
{
	return catalytic_atoms_geometry_penalty;
}

/**
 * returns the relative SASA penalty score of this motif hit
 */
const double MotifHit::get_relative_SAS_penalty() const
{
	return relative_SAS_penalty;
}

/**
 * returns the backbone clash penalty score of this motif hit
 */
const double MotifHit::get_backbone_clash_penalty() const
{
	return backbone_clash_penalty;
}

/**
 * returns the spheres clash penalty score of this motif hit
 */
const double MotifHit::get_ligand_clash_penalty() const
{
	return ligand_clash_penalty;
}

/**
 * returns the interaction center penalty score of this motif hit
 */
const double MotifHit::get_interaction_centers_penalty() const
{
	return interaction_centers_penalty;
}

/**
 * returns a unique identifier string for the motif hit. it consists of the name of the PDB structure,
 * the ID and Chain ID of each anchor residue of the motif hit in the same order as the residues in the motif
 * definition file / motif object
 */
const String MotifHit::get_ID() const
{
	String motif_hit_id = pdb_name;
	for (unsigned int j = 0; j < motif.get_residue_count(); ++j)
	{
		motif_hit_id += " " + all_residues[indices[j]]->getID();
		if (all_residues[indices[j]]->getChain()->getName() != " ")
		{
			motif_hit_id += ":" + all_residues[indices[j]]->getChain()->getName();
		}
	}
	return motif_hit_id;

}

const String MotifHit::get_rotamer_indices() const
{
    String res;
    for (vector<unsigned int>::const_iterator it = rotamer_indices.begin();
              it != rotamer_indices.end(); ++it)
        res += String((*it + 1)) + " ";
    return res;
}

void MotifHit::set_pair_rotamer_score(unsigned int pair_index,
                            unsigned int rot1, unsigned int rot2, double score)
{
	pair_rotamer_scores[pair_index][rot1*10000 + rot2] = score;
}

/**
 * contains the vdW-Radius of an oxygen atom
 */
const double MotifHit::oxygen_radius = 1.52;

/**
 * contains the vdW-Radius of a carbon atom
 */
const double MotifHit::carbon_radius = 1.7;

/**
 * contains the vdW-Radius of a nitrogen atom
 */
const double MotifHit::nitrogen_radius = 1.55;

/**
 * contains the vdW-Radius of a sulfur atom
 */
const double MotifHit::sulfur_radius = 1.85;
