/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "scaffoldSearch.h"
using namespace BALL;
using namespace std;

/*
 * Constructor that reads in the parameters for the scaffold search and stores them in the
 * corrsponding memeber variables. stores the motif in the memeber variabel motif
 */
ScaffoldSearch::ScaffoldSearch(Parameters const &parameter_file,Motif& motif_ref) :
	parameters( parameter_file),
	ranking(),
	file_list(),
	motif( motif_ref)
{
    if (motif.get_interaction_center_count() == 0)
    {
    	parameters.set_calculate_interaction_centers(false);
    }
    ranking = Ranking(parameters, motif.get_name());

}


ScaffoldSearch::~ScaffoldSearch()
{
}


void ScaffoldSearch::subdirectory_recursion(vector<String>& files, HashSet<String> const &include_list, String& path)
{
    DIR *dp;
    struct dirent *dirp;
    if((dp  = opendir((parameters.get_structure_directory() + "/" + path).c_str())) == NULL)
    {
       	Exception::GeneralException e("parameter file", -1, "invalid NoDirectory",
       	                              parameters.get_structure_directory() + "/" + path + " cannot be opened.");
       	throw e;
    }

    while ((dirp = readdir(dp)) != NULL)
    {
    	String filename = String(dirp->d_name);
    	if (filename == "." || filename == "..") continue;
    	filename = filename.trim();
    	String filepath =  path + "/" + filename;
    	if (!chdir((parameters.get_structure_directory() + filepath).c_str()) )
    	{
    		subdirectory_recursion(files, include_list, filepath);
    	}
    	else
    	{
    		if (include_list.has(filename))
    		{
    			files.push_back(filepath);
    		}
    	}
    }
    closedir(dp);
}

/*
 * returns all filenames in the directory with the name dir
 * and it's subdirectories
 */
vector<String> ScaffoldSearch::create_pdb_list ()
{
	HashSet<String> include_list_set;

	try
	{
		LineBasedFile f;
    	f.open(parameters.get_include_list_file());
    	String line;

		// read all lines of the parameter file
	    while (f.readLine())
	    {
	    	line = f.getLine();
	    	line = line.trim();
	    	include_list_set.insert(line);
	    }
	}
	catch (Exception::FileNotFound &e)
	{
		cerr << "ERROR encountered: "
			 << endl <<"The include list file \"" << e.getFilename()
			 << "\" could not be found" << endl;
	}
	vector<String> files;
	String s("");
    subdirectory_recursion(files, include_list_set, s);

    return files;
}


/*
 * starts a search for suitable scaffolds in all pdb-files in the file list
 */
void ScaffoldSearch::start_scaffold_search()
{
    // prepare the list of pdb files to search
    file_list = create_pdb_list();

    if (file_list.size() == 0)
    {
    	cout << "No PDB File listed in \"" << parameters.get_include_list_file()
    	     << "\" found in directory \"" << parameters.get_structure_directory() << "\"." << endl;
    	return;
    }

	vector<String>::iterator fit = file_list.begin();

	unsigned int motif_count = 0;

	Timer timer;
	double time = 0.0, time_since_last_ranking = 0.0;
	// loop over all PDB files in the list
	for (; fit != file_list.end(); ++fit)
	{
		cout << "Processing File \"" << *fit << "\"" << endl;
		try
		{
			timer.start();
			System system = System();
			// give the PDB system the name of the PDB file
			system.setName(*fit);
			PDBFile pdb_file;
			pdb_file.open(parameters.get_structure_directory() + *fit);
			// if the file has the wrong format, skip it
			if(!pdb_file.hasFormat())
			{
				Exception::GeneralException e(__FILE__, __LINE__,
				                      "Unsuitable PDB file",
				                      "File "+ *fit + "is no valid PDB file.");
				throw e;
			}

			pdb_file.read(system);

			// if the file has the wrong format, skip it
			if(!system.isValid())
			{
				Exception::GeneralException e(__FILE__, __LINE__, "Unsuitable PDB file", "File "+ *fit + " cannot be read as a valid PDB System.");
				throw e;
			}

			// remove all non-amino acid residues from the system
			// so the number of residues only counts amino acids and the SAS areas of
			// surface residues are not altered by water molecules etc in the PDB file
			ResidueIterator res_it = system.beginResidue();
			for (; res_it != system.endResidue(); ++ res_it)
			{
				if (!res_it->isAminoAcid())
				{
					res_it->select();
				}
			}
			system.removeSelected();

			// if the protein has too few or too many residues, skip it
			if (system.countResidues() < parameters.get_minimal_residue_number())
			{
				Exception::GeneralException e(__FILE__, __LINE__, "Unsuitable PDB file", "Structure "+ *fit + " does not contain enough residues.");
				throw e;
			}

			else if (system.countResidues() > parameters.get_maximal_residue_number())
			{
				Exception::GeneralException e(__FILE__, __LINE__, "Unsuitable PDB file", "Structure "+ *fit + " contains too many residues.");
				throw e;
			}

			LineBasedFile file;
			file.open(parameters.get_structure_directory() + "/" + *fit);
			String line;
			bool remark = false;
			bool have_one = false;
			// loop over the lines of the PDB file to determine the resolution of the structure
			while (file.readLine())
			{
				line = file.getLine();

				// the resolution should be stored in a line of the format:
				// REMARK   2 RESOLUTION. 2.3  ANGSTROMS.
				if (line.getField(0) == "REMARK")
				{
					remark = true;
					if (line.getField(2) == "RESOLUTION.")
					{
						if (line.getField(3).isFloat())
						{
							if (line.getField(3).toDouble() > parameters.get_minimal_resolution())
							{
								Exception::GeneralException e(__FILE__, __LINE__, "Unsuitable PDB file", "Structure "+ *fit + " has a resolution that is too low.");
								throw e;
							}
							else
							{
								if (have_one) break;
								else have_one = true;
							}
						}
						{
							if (have_one) break;
							else have_one = true;
						}
					}
					// REMARK 200  EXPERIMENT TYPE                : X-RAY DIFFRACTION
					if (line.getField(2) == "EXPERIMENT" && line.getField(3) == "TYPE")
					{
						if ( !(line.getField(5) == "X-RAY" || line.getField(6) == "X-RAY") )
						{
							Exception::GeneralException e(__FILE__, __LINE__, "Unsuitable PDB file", "Structure "+ *fit + " was not solved by X-ray diffraction.");
							throw e;
						}
						else
						{
							if (have_one) break;
							else have_one = true;
						}
					}
				}
				// we checked all lines in the "REMARK" section, if the reoslution wasnt entered there, stop the search
				// the structrue will be searched for motif hits just in case
				else if (remark)
				{
					break;
				}
			}


			// create a MotifSearch object for the PDB structure
			// with it we will search for motif hits in the structure and compute all
			// penalties and scores for the hits that are flagged "true" in the parameter file
			MotifSearch motif_search(system, motif, parameters);

			// start the search and add up the found motif hits
			motif_count += motif_search.search_for_motifs(ranking);

			timer.stop();
			time_since_last_ranking += timer.getUserTime();

			if (time_since_last_ranking > ranking_time_interval)
			{
				time_since_last_ranking = 0.0;
				ranking.write_rankings();
				cout << "Preliminary ranking written. " << endl;
			}

			if (parameters.time_is_of_the_essence)
			{
				cout << "Complete time for this PDB file: " << timer.getUserTime() << " s" << endl;
				time += timer.getUserTime();
			}
			timer.reset();
		}
		catch (Exception::GeneralException &e)
		{
			if (String(e.getName()) == "Unsuitable PDB file")
			{
				cout << e.getMessage() << endl;
			}
			else throw e;
		}
	}
	// write out the ranking lists for the different scores/penalties of the motif hits
	ranking.write_rankings ();
	cout << "Done.  " << motif_count << " motif attachment sites found. ";
	if (parameters.time_is_of_the_essence)
	    cout << "Average time per PDB file: " << time/file_list.size() << endl;
	cout << endl;

}

