/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "scaffold_selection_utilities.h"
using namespace BALL;
using namespace std;


/*constructs the position of the HA1 atom, which is used as a
 * CB subsitute, by assuming an ideal tetrahedral geometry of the atoms
 * N, C, CA and HA1
 */
Vector3 scsel_util::get_ideal_cb_position_for_gly (Vector3 const &N, Vector3 const &C, Vector3 const &CA)
{
   	// Ideal positions for the atoms of Glycin that will be used for the computation of HA1 position
   	// as a CB substitute. The ideal positions are derived of a tetrahedral geometry and a bond length of 1.54 A
   	Vector3 idealca, idealn, idealha1, idealc;
   	idealca = Vector3(0.889119415, 0.889119415, 0.889119415);
	idealn = Vector3(1.778238829, 0.0, 0.0);
	idealha1 = Vector3(1.778238829, 1.778238829, 1.778238829);
	idealc = Vector3(0.0, 0.0, 1.778238829);

   	Matrix4x4	T;
	StructureMapper	mapper;

	// compute the transformation Matrix that maps the ideal positions of the Backbone atoms onto the
	// ones from the structure; thaen use the matrix to compute the ideal position of the HA1 Atom in
	// the structure
	T = mapper.matchPoints(idealca, idealn, idealc, CA, N, C);
	Vector3 ideal_CB = T * idealha1;
	return ideal_CB;
}

pair <Vector3, Vector3> scsel_util::get_rotamer_ca_positions_for_ser_cys (Vector3 const &CA, Vector3 const &CB, Vector3 const &O_S_G, bool is_serine)
{
   	// Ideal positions for the atoms of Ser/Cys that will be used for the computation of the two other CA positions
   	// as a CB substitute. The ideal positions are derived of a tetrahedral geometry and a bond length of 1.54 A for the CA-CB bond
   	Vector3 CA_template(0.0, 0.0, 0.0), CB_template, O_S_G_template, CA1_template, CA2_template;
   	if (is_serine)
   	{
   		CB_template = Vector3(0.539,   1.178,  -0.820);
   		O_S_G_template = Vector3(1.941,   1.085,  -0.979);
   		CA1_template = Vector3(0.174,   2.494,  -0.130);
   		CA2_template = Vector3(-0.117,   1.151,  -2.202);
   	}
   	else
   	{
   		CB_template = Vector3(0.588,   1.170,  -0.792);
   		O_S_G_template = Vector3(2.385,   1.123,  -0.591);
   		CA1_template = Vector3(0.235,   1.039,  -2.275);
   		CA2_template = Vector3(0.055,   2.500,  -0.257);
   	}

   	Matrix4x4	T;
	StructureMapper	mapper;

	// compute the transformation Matrix that maps the ideal positions of the Backbone atoms onto the
	// ones from the structure; thaen use the matrix to compute the ideal position of the HA1 Atom in
	// the structure
	T = mapper.matchPoints(O_S_G_template, CB_template, CA_template, O_S_G, CB, CA);
	Vector3 CA1 = T * CA1_template;
	Vector3 CA2 = T * CA2_template;
	return pair<Vector3, Vector3>(CA1, CA2);
}


vector<double> scsel_util::compute_geometric_values (Vector3 const &CA1, Vector3 const &CB1,
                                                   Vector3 const &CA2, Vector3 const &CB2)
{
	vector<double> geometric_values(4);
  	Vector3 d1 = CA1 - CA2;
  	Vector3 d2 = CA2 - CA1;

  	// compute Ca distance
  	double dist = d1.getLength();

    // compute angle Cb1 - Ca1 - Ca2
    double phi  = (CB1 - CA1).getAngle(d1).toDegree();

    // compute angle Ca1 - Ca2 - Cb2
    double psi = (CB2 - CA2).getAngle(d2).toDegree();

    // compute dihedral angle Cb1 - Ca1 - Ca2 - Cb2
	double tau = getTorsionAngle(CB1[0], CB1[1], CB1[2], CA1[0], CA1[1], CA1[2],
	                             CA2[0], CA2[1], CA2[2], CB2[0], CB2[1], CB2[2]).toDegree();

	geometric_values[0] = dist;
	geometric_values[1] = phi;
	geometric_values[2] = psi;
	geometric_values[3] = tau;
	return geometric_values;
}

/*
 * returns the corresponding amino acid one letter code for the three letter code argument
 */
String scsel_util::get_one_letter_code(String const &three_letter_code)
{
	if (three_letter_code == "HIS")
		return "H";
	if (three_letter_code ==  "ASP")
		return "D";
	if (three_letter_code ==  "SER")
		return "S";
	if (three_letter_code ==  "GLU")
		return "D";
	if (three_letter_code ==  "GLN")
		return "Q";
	if (three_letter_code ==  "TRP")
		return "W";
	if (three_letter_code ==  "TYR")
		return "Y";
	if (three_letter_code ==  "LYS")
		return "K";
	if (three_letter_code ==  "ARG")
		return "R";
	if (three_letter_code ==  "THR")
		return "T";
	if (three_letter_code ==  "PHE")
		return "F";
	if (three_letter_code ==  "ILE")
		return "I";
	if (three_letter_code ==  "LEU")
		return "L";
	if (three_letter_code ==  "CYS")
		return "C";
	if (three_letter_code ==  "MET")
		return "M";
	if (three_letter_code ==  "HIS")
		return "H";
	if (three_letter_code ==  "VAL")
		return "V";
	if (three_letter_code ==  "GLY")
		return "G";
	if (three_letter_code ==  "ALA")
		return "A";
	if (three_letter_code ==  "PRO")
		return "P";
	String message = "\"" + three_letter_code + "\" is no valid three letter amino acid code";
	Exception::GeneralException e;
	e.setMessage(message);
	throw e;
}

vector<double> scsel_util::read_normal_distribution_vector(const String &filename)
{
	vector<double> normal_distribution_vector;
	vector<double> column_heads;
    LineBasedFile f;
    f.open(filename);
    String line;
    unsigned int linecount = 1;
    try
    {
	    while(f.readLine())
	    {
	    	line = f.getLine();
	    	if (line.getField(0).compare("z") == 0)
	    	{
	    		column_heads.push_back(0.0);
	    		for (unsigned int i = 1; i < line.countFields(); ++i)
	    		{
					column_heads.push_back(line.getField(i).toDouble());
				}
	    	}
	    	else
	    	{
	    		for (unsigned int i = 1; i < line.countFields(); ++i)
	    		{
					normal_distribution_vector.push_back(line.getField(i).toDouble());
				}
	    	}
	    	linecount++;
	    }
    }
    catch (...)
    {
    	Exception::GeneralException e((filename.c_str()), linecount, "InvalidFormat",
    	                            "normal distribution table file has invalid format");
    	throw e;
    }
    f.close();

    //  the vector now contains 350 probabilities for values from 0.0 to 3.49 .
    //  the probability for value x is in vector index x*100
    return normal_distribution_vector;
}
